/*
** $Id: audio-entropyd.c,v 1.1.1.1 1999/08/15 14:04:56 dmiller Exp $
**
** Simple program to reseed kernel random number generator using
** data read from soundcard.
**
** Copyright 1999 Damien Miller <djm@mindrot.org>
**
** This code is licensed under the GNU Public License version 2
** Please see the file COPYING for more details.
**
*/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>

#include <linux/soundcard.h>
#include <asm/types.h>
#include <linux/random.h>

#include "sha1.h"

#define DEFAULT_DSP_DEV					"/dev/dsp"
#define DEFAULT_MIXER_DEV				"/dev/mixer"
#define DEFAULT_SLEEP_TIME				300000000
#define RANDOM_DEVICE					"/dev/random"
#define URANDOM_DEVICE					"/dev/urandom"
#define DEFAULT_INPUT_BUFFER_SIZE	65536
#define DEFAULT_SEED_BUFFER_SIZE		60
#define DEFAULT_CREDIT					320
#define DEFAULT_SAMPLE_RATE			44100
#define PID_FILE							"/var/run/audio-entropyd.pid"

/* Prototypes */
void main_loop(const char *dsp_device, int sleep_time, int update_krng, 
					const char *raw_output, const char *hash_output, 
					int read_size, int seed_size, int credit, int sample_rate);
void setup_mixer(const char *mixer_device);
int setup_dsp_device(const char *dsp_device, int sample_rate);
void usage(void);
void urandom_bytes(char *s, size_t len);
void credit_krng(int random_fd, struct rand_pool_info *entropy);
void build_entropy_info(struct rand_pool_info *entropy, int seed_size, 
								int credit, unsigned char *diff_buffer, int diff_size);
void daemonise(void);
void gracefully_exit(int signum);

/* Functions */

int main(int argc, char **argv)
{
	char *dsp_device = DEFAULT_DSP_DEV;
	char *mixer_device = DEFAULT_MIXER_DEV;
	char *raw_output = NULL;
	char *hash_output = NULL;
	int read_size = DEFAULT_INPUT_BUFFER_SIZE;
	int seed_size = DEFAULT_SEED_BUFFER_SIZE;
	int credit = DEFAULT_CREDIT;
	int sample_rate = DEFAULT_SAMPLE_RATE;
	int sleep_time = DEFAULT_SLEEP_TIME;
	int update_krng = 1;
	int c;
	static struct option long_options[] =
	{
		{"read-size",		1, NULL, 'r'},
		{"seed-size",		1, NULL, 's'},
		{"credit",			1, NULL, 'c'},
		{"dsp-device",		1, NULL, 'd'},
		{"mixer-device",	1, NULL, 'm'},
		{"wait",				1, NULL, 'w'},
		{"raw-output",		1, NULL, 'R'},
		{"seed-output",	1, NULL, 'S'},
		{"sample-rate",	1, NULL, 'N'},
		{"no-mixer",		0, NULL, 'n'},
		{"no-update",		0, NULL, 'K'},
		{"help",				0, NULL, 'h'},
		{NULL,				0, NULL, 0}
	};

	/* Process commandline options */
	while(1)
	{
		c = getopt_long (argc, argv, "r:s:c:d:m:w:R:S:N:nKh", long_options, NULL);
		if (c == -1)
			break;
		
		switch(c)
		{
			case 'r':
				read_size = atoi(optarg);
				break;
			case 's':
				seed_size = atoi(optarg);
				break;
			case 'c':
				credit = atoi(optarg);
				break;
			case 'd':
				dsp_device = strdup(optarg);
				break;
			case 'm':
				mixer_device = strdup(optarg);
				break;
			case 'w':
				sleep_time = atoi(optarg);
				break;
			case 'R':
				raw_output = strdup(optarg);
				break;
			case 'S':
				hash_output = strdup(optarg);
				break;
			case 'N':
				sample_rate = atoi(optarg);
				break;
			case 'n':
				mixer_device = NULL;
				break;
			case 'K':
				update_krng = 0;
				break;
			case 'h':
				usage();
				exit(0);
			default:
				fprintf(stderr, "Invalid commandline options.\n\n");
				usage();
				exit(1);
		}
	}
	
	if (read_size == 0)
	{
		fprintf(stderr, "Error: Read size must be a greater than zero.\n");
		exit(1);
	}
	if (read_size & (read_size - 1))
	{
		fprintf(stderr, "Error: Read size must be a power of two.\n");
		exit(1);
	}
	if (seed_size == 0)
	{
		fprintf(stderr, "Error: Seed size must be a greater than zero.\n");
		exit(1);
	}
	if ((seed_size % SHA1_HASH_LEN) != 0)
	{
		fprintf(stderr, "Error: Seed size must be a multiple of %i.\n", SHA1_HASH_LEN);
		exit(1);
	}
	if (seed_size >= read_size)
	{
		fprintf(stderr, "Error: Seed size must be less than read size.\n");
		exit(1);
	}
	if (credit > (seed_size * 8))
	{
		fprintf(stderr, "Error: Credit must be less than (8 * seed size).\n");
		exit(1);
	}

	daemonise();

	signal(SIGHUP, gracefully_exit);
	signal(SIGINT, gracefully_exit);
	signal(SIGTERM, gracefully_exit);

#ifdef DEBUG
	openlog("audio-entropyd", LOG_CONS|LOG_PERROR, LOG_DAEMON);
#else /* !DEBUG */
	openlog("audio-entropyd", LOG_CONS, LOG_DAEMON);
#endif /* DEBUG */

	syslog(LOG_NOTICE, "audio-entropyd starting up");

	if (mixer_device != NULL)
		setup_mixer(mixer_device);

	main_loop(dsp_device, sleep_time, update_krng, raw_output, 
				 hash_output, read_size, seed_size, credit, sample_rate);
	
	exit(0);
}

int setup_dsp_device(const char *dsp_device, int sample_rate)
{
	int dsp_fd;
	int setting;

	/* Open sound device */
	dsp_fd = open(dsp_device, O_RDONLY);
	if (dsp_fd == -1)
	{
		syslog(LOG_CRIT, "Could not open %s for reading: %m", dsp_device);
		exit(1);
	}

	/* Set sample rate */
	if (ioctl(dsp_fd, SNDCTL_DSP_SPEED, &sample_rate) == -1)  
	{
		syslog(LOG_CRIT, "Could not set %i sample rate: %m", setting);
		exit(1);
	}
	
	/* Set sample format */
	setting = AFMT_S16_LE; /* 16 bit little endian */
	if (ioctl(dsp_fd, SNDCTL_DSP_SETFMT, &setting) == -1)
	{
		syslog(LOG_CRIT, "Could not set 16 bit sample mode: %m");
		exit(1);
	}
	
	/* Set stereo */
	setting = 1; /* Stereo */
	if ((ioctl(dsp_fd, SNDCTL_DSP_STEREO, &setting) == -1) || (setting != 1))
	{
		syslog(LOG_CRIT, "Could not set stereo sample mode: %m");
		exit(1);
	}
	
	return(dsp_fd);
}

void setup_mixer(const char *mixer_device)
{
	int mixer_fd;
	int setting;
	int input_channels;
	int c;
#ifdef DEBUG
	const char *labels[] = SOUND_DEVICE_LABELS;
#endif

	/* Open mixer device */
	mixer_fd = open(mixer_device, O_RDWR);
	if (mixer_fd == -1)
	{
		syslog(LOG_WARNING, "Could not open %s for reading: %m", mixer_device);
		syslog(LOG_WARNING, "Disabling mixer support, output quality may be degraded");
		return;
	}
	
	/* Query for input channels */
	if (ioctl(mixer_fd, SOUND_MIXER_READ_RECMASK, &input_channels) == -1)
	{
		syslog(LOG_WARNING, "Could not get recoding mask from mixer: %m");
		syslog(LOG_WARNING, "Disabling mixer support, output quality may be degraded");
		return;
	}

	/* Any channels available? */
	if (input_channels == 0)
	{
		syslog(LOG_CRIT, "No input channels available");
		exit(1);
	}

	/* Set each input channel to maximum volume */
	for(c = 0; c < SOUND_MIXER_NRDEVICES; c++)
	{
		/* Is an input channel? */
		if (input_channels & 1)
		{
#ifdef DEBUG
			syslog(LOG_DEBUG, "Setting volume to max on channel %i (%s)", c, labels[c]);
#endif		
			/* If so, set volume */
			setting = 100 + (100 << 8);
			if (ioctl(mixer_fd, MIXER_WRITE(c), &setting) == -1)
				syslog(LOG_WARNING, "Could not set mixer channel %i: %m", c);
		}
		
		/* Move on to next channel */
		input_channels >>= 1;
	}
	
	close(mixer_fd);
}

void main_loop(const char *dsp_device, int sleep_time, int update_krng, 
					const char *raw_output, const char *hash_output, 
					int read_size, int seed_size, int credit, int sample_rate)
{
	unsigned char *input_buffer;
	unsigned char *diff_buffer;

	int bytes_read;
	int c;
	int d;

	int dsp_fd = -1;
	int random_fd = -1;
	int raw_fd = -1;
	int hash_fd = -1;

	struct rand_pool_info *entropy;

	input_buffer = malloc(read_size);
	diff_buffer = malloc(read_size / 2);

	entropy = malloc(sizeof(struct rand_pool_info) + seed_size);

	if ((input_buffer == NULL) || (diff_buffer == NULL) || (entropy == NULL))
	{
		syslog(LOG_CRIT, "malloc failed");
		exit(1);
	}

	if (raw_output != NULL)
	{
		raw_fd = open(raw_output, O_WRONLY|O_CREAT|O_TRUNC, 0600);
		if (raw_fd == -1)
		{
			syslog(LOG_CRIT, "Couldn't open raw output file: %m");
			exit(1);
		}
	}

	if (hash_output != NULL)
	{
		hash_fd = open(hash_output, O_WRONLY|O_CREAT|O_TRUNC, 0600);
		if (hash_fd == -1)
		{
			syslog(LOG_CRIT, "Couldn't open hash output file: %m");
			exit(1);
		}
	}

	/* Open kernel random device */
	random_fd = open(RANDOM_DEVICE, O_RDWR);
	if (random_fd == -1)
	{
		syslog(LOG_CRIT, "Couldn't open random device: %m");
		exit(1);
	}
	
	/* Prepopulate hash buffer with random data */
	/* If attacker somehow manages to take total control of soundcard */
	/* the bytes entering the KRNG are still somewhat unknown */
	/* hash buffer is mixed with data coming from soundcard later */
	urandom_bytes((unsigned char *)entropy->buf, seed_size);
	
	/* Open and set up DSP for reading */
	dsp_fd = setup_dsp_device(dsp_device, sample_rate);	

	/* Discard the first data read */
	/* it often contains weird looking data - probably a click from */
	/* driver loading / card initialisation */
	read(dsp_fd, input_buffer, 8192);

	/* Main read loop */
	while (1)
	{
		/* Read a buffer of audio */
		bytes_read = read(dsp_fd, input_buffer, read_size);
		if ((bytes_read == -1) || (bytes_read != read_size))
		{
			syslog(LOG_CRIT, "Read error: %m");
			exit(1);
		}

		/* Take difference between left and right channels */
		/* this should cancel some external signals */
		d = 0;
		for(c = 0; c < (read_size / 2); c += 2)
		{
			diff_buffer[c]     = input_buffer[d]     - input_buffer[d + 2];
			diff_buffer[c + 1] = input_buffer[d + 1] - input_buffer[d + 3];
			d += 4;
		}

		/* Build struct to pass to kernel */
		build_entropy_info(entropy, seed_size, credit, diff_buffer, read_size / 2);

		if (raw_output != NULL)
		{
			if (write(raw_fd, diff_buffer, read_size / 2) != (read_size / 2)) 
			{
				syslog(LOG_CRIT, "Write error on raw output file: %m");
				exit(1);
			}
		}

		if (hash_output != NULL)
		{
			if (write(hash_fd, (unsigned char *)entropy->buf, 40) != 40)
			{
				syslog(LOG_CRIT, "Write error on hash output file: %m");
				exit(1);
			}
		}

		if (update_krng)
			credit_krng(random_fd, entropy);
		
		/* Diddle the sleep time a little bit to avoid syncronisation */
		/* with periodic signals. <<7 gives up to 32ms shift which will */
		/* deal with signals with frequencies greater than 31.25hz */
		usleep(sleep_time + (((unsigned char *)entropy->buf)[0] << 6));
	}
}

void usage(void)
{
	fprintf(stderr, "Usage: audio-entropyd [options]\n\n");
	fprintf(stderr, "Collect entropy from a soundcard and feed it into the kernel random pool.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "--read-size,    -r []  Number of bytes to read at a time. (Default %i)\n", DEFAULT_INPUT_BUFFER_SIZE);
	fprintf(stderr, "--seed-size,    -s []  Number of bytes in final hash. (Default %i)\n", DEFAULT_SEED_BUFFER_SIZE);
	fprintf(stderr, "--credit,       -c []  Number of bits of seed to credit to kernel. (Default %i)\n", DEFAULT_CREDIT);
	fprintf(stderr, "--dsp-device,   -d []  Specify sound device to use. (Default %s)\n", DEFAULT_DSP_DEV);
	fprintf(stderr, "--mixer-device, -m []  Specify mixer device to use. (Default %s)\n", DEFAULT_MIXER_DEV);
	fprintf(stderr, "--wait,         -w []  How long (us) to wait between reads. (Default %i)\n", DEFAULT_SLEEP_TIME);
	fprintf(stderr, "--raw-output,   -R []  Write raw differenced audio to file (for testing).\n");
	fprintf(stderr, "--seed-output,  -S []  Write seed material to file (for testing).\n");
	fprintf(stderr, "--sample-rate,  -N []  Audio sampling rate. (default %i)\n", DEFAULT_SAMPLE_RATE);
	fprintf(stderr, "--no-mixer,     -n     Do not reset mixer\n");
	fprintf(stderr, "--no-update,    -K     Do not update kernel RNG (good for testing)\n");
	fprintf(stderr, "\n");
}

void urandom_bytes(char *s, size_t len)
{
	int 		c;
	int		rng_fd;
	
	rng_fd = open(URANDOM_DEVICE, O_RDONLY);
	if (rng_fd == -1)
	{
		syslog(LOG_ERR, "Couldn't open urandom device for reading: %m");
		exit(1);
	}

	c = read(rng_fd, s, len);
	if (c != len)
	{
		close(rng_fd);
		syslog(LOG_ERR, "Couldn't read from /dev/urandom: %m");
		exit(1);
	}
	
	close(rng_fd);
}

void credit_krng(int random_fd, struct rand_pool_info *entropy)
{
	int before;
	int after;

	/* Get number of bits in KRNG before credit */
	if (ioctl(random_fd, RNDGETENTCNT, &before) == -1)
	{
		syslog(LOG_CRIT, "Coundn't query entropy to kernel: %m");
		exit(1);
	}

	/* Dump it in the kernel random pool */
	if (ioctl(random_fd, RNDADDENTROPY, entropy) == -1)
	{
		syslog(LOG_CRIT, "Coundn't credit entropy to kernel: %m");
		exit(1);
	}

	/* Get number of bits in KRNG after credit */
	if (ioctl(random_fd, RNDGETENTCNT, &after) == -1)
	{
		syslog(LOG_CRIT, "Coundn't query entropy to kernel: %m");
		exit(1);
	}

	syslog(LOG_INFO, "Entropy credit of %i/%i made (%i bits before, %i bits after)", 
			 entropy->entropy_count, entropy->buf_size * 8, before, after);
}

void build_entropy_info(struct rand_pool_info *entropy, int seed_size, 
								int credit, unsigned char *diff_buffer, int diff_size)
{
	int num_hashes;
	int diff_bytes_per_hash;
	int c;
	int diff_offset;
	int hash_offset;
	SHA1_CTX context;

	num_hashes = seed_size / SHA1_HASH_LEN;
	diff_bytes_per_hash = diff_size / num_hashes;
	
	/* Set up entropy structure for credit */
	entropy->entropy_count = credit;
	entropy->buf_size = seed_size;

	/* Take sha1 hashes of the data, put results in entropy buffer */
	/* Hash in contents of old seed buffer too */
	diff_offset = 0;
	hash_offset = 0;
	for(c = 0; c < num_hashes; c++)
	{
		SHA1_Init(&context);
		SHA1_Update(&context, diff_buffer + diff_offset, diff_bytes_per_hash);
		SHA1_Update(&context, (unsigned char *)(entropy->buf) + hash_offset, SHA1_HASH_LEN);
		SHA1_Final((unsigned char *)(entropy->buf) + hash_offset, &context);
		
		diff_offset += diff_bytes_per_hash;
		hash_offset += SHA1_HASH_LEN;
	}
}

void daemonise(void)
{
	FILE *p;
	
	switch (fork())
	{
		case -1:
			syslog(LOG_CRIT, "Couldn't fork.");
			fprintf(stderr, "Couldn't fork.\n");
			exit(2);
		case 0:
			/* We are the child */
			break;
		default:
			/* We are the parent */
			_exit(0);
	}

	setsid();

	switch (fork())
	{
		case -1:
		syslog(LOG_CRIT, "Couldn't fork.");
		exit(2);
	case 0:
		/* We are the child */
		break;
	default:
		/* We are the parent */
		_exit(0);
	}

	chdir("/");

	umask(077);

	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	p = fopen(PID_FILE, "w");
	if (p == NULL)
	{
		syslog(LOG_CRIT, "Couldn't open PID file \"%s\" for writing: %m.", PID_FILE);
		exit(2);
	}
	fprintf(p, "%i", getpid());
	fclose(p);
}

void gracefully_exit(int signum)
{
	unlink(PID_FILE);
	syslog(LOG_NOTICE, "audio-entropyd stopping");
	exit(0);
}
